@@ -1,3 +1,15 @@
+/*
+ * The modifications to support SSLeay were done by Tim Hudson
+ * tjh@cryptsoft.com and Christoph Martin martin@uni-mainz.de
+ *
+ * You can do whatever you like with these patches except pretend that
+ * you wrote them.
+ *
+ * Email ssl-users-request@mincom.oz.au to get instructions on how to
+ * join the mailing list that discusses SSLeay and also these patches.
+ *
+ */
+
 /*
  * Copyright (c) 1988, 1990 Regents of the University of California.
  * All rights reserved.
@@ -35,7 +47,7 @@
  * From: @(#)telnet.c	5.53 (Berkeley) 3/22/91
  */
 char telnet_rcsid[] = 
-"$Id: telnet.cc,v 1.34 1999/08/19 09:34:15 dholland Exp $";
+"$Id: telnet.cc,v 1.36 2000/07/23 03:24:53 dholland Exp $";
 
 #include <string.h>
 #include <sys/types.h>
@@ -67,6 +79,10 @@ char telnet_rcsid[] =
 #include <term.h>
 #endif
 
+#if defined(AUTHENTICATE)
+#include <libtelnet/auth.h>
+#include <libtelnet/misc-proto.h>
+#endif
 
 #define	strip(x)	((x)&0x7f)
 
@@ -178,6 +194,10 @@ void init_telnet(void) {
   memset(options, 0, sizeof(options));
   
   connected = In3270 = ISend = localflow = donebinarytoggle = 0;
+
+#if	defined(ENCRYPT) || defined(AUTHENTICATE)
+  auth_encrypt_connect(connected);
+#endif
   
   SYNCHing = 0;
   
@@ -339,6 +359,9 @@ void willoption(int option) {
       settimer(modenegotiated);
       /* FALL THROUGH */
     case TELOPT_STATUS:
+#if	defined(AUTHENTICATE)
+    case TELOPT_AUTHENTICATION:
+#endif
       new_state_ok = 1;
       break;
       
@@ -399,6 +422,21 @@ void wontoption(int option) {
       set_my_state_dont(option);
       return;		/* Never reply to TM will's/wont's */
       
+#ifdef USE_SSL
+      /* don't know if this is needed here ... don't
+       * really want to think about it myself
+       * at the moment --tjh 
+       */
+    case TELOPT_AUTHENTICATION:
+      if (ssl_secure_flag && !ssl_only_flag) {
+	fflush(stderr);
+	printf("[SSL not available]\n");
+	fflush(stdout);
+	/* abort quickly ... */
+	exit(1);
+      }
+#endif /* USE_SSL */
+
     default:
       break;
     }
@@ -454,9 +492,15 @@ static void dooption(int option) {
       case TELOPT_ENVIRON:	/* environment variable option */
 	new_state_ok = 1;
 	break;
+#if	defined(AUTHENTICATE)
+      case TELOPT_AUTHENTICATION:
+	if (autologin)
+	  new_state_ok = 1;
+	break;
+#endif
 	
       case TELOPT_XDISPLOC:	/* X Display location */
-	if (env_getvalue("DISPLAY"))
+	if (env_getvalue("DISPLAY", 0))
 	  new_state_ok = 1;
 	break;
 	
@@ -519,6 +563,19 @@ static void dontoption(int option) {
     case TELOPT_LINEMODE:
       linemode = 0;	/* put us back to the default state */
       break;
+#ifdef USE_SSL
+    case TELOPT_AUTHENTICATION:
+      if (ssl_secure_flag && !ssl_only_flag) {
+	fflush(stderr);
+	printf("[SSL not available]\n");
+	fflush(stdout);
+	/* abort quickly ... */
+	exit(1);
+      } else {
+	fprintf(stderr,"[SSL not available]\n");
+	fflush(stderr);
+      }
+#endif /* USE_SSL */
     }
     /* we always accept a DONT */
     set_my_want_state_wont(option);
@@ -600,7 +657,9 @@ static void mklist(char *buf, const char *name, stringarray &fill) {
    */
   
   fill.add(fill[fill.num()-1]);
-  fill.add(NULL);
+
+  /* dholland 21-May-2000 I think this is bogus */
+  /*fill.add(NULL);*/
 }
 
 char termbuf[2048];
@@ -636,7 +695,7 @@ static const char *gettermname(void) {
   
   if (resettermname) {
     resettermname = 0;
-    tname = env_getvalue("TERM");
+    tname = env_getvalue("TERM", 0);
     if (!tname || my_setupterm(tname, 1, &err)) {
       termbuf[0] = 0;
       tname = "UNKNOWN";
@@ -644,7 +703,7 @@ static const char *gettermname(void) {
     mklist(termbuf, tname, termtypes);
     next = 0;
   }
-  if (next==termtypes.num()) next = 0;
+  if (next==termtypes.num()-1) next = 0;
   return termtypes[next++];
 }
 /*
@@ -679,7 +738,7 @@ static void suboption(void) {
       }
 #endif /* TN3270 */
       name = gettermname();
-      netoring.printf("%c%c%c%c%s%c%c", IAC, SB, TELOPT_TTYPE,
+      netoring.xprintf("%c%c%c%c%s%c%c", IAC, SB, TELOPT_TTYPE,
 		      TELQUAL_IS, name, IAC, SE);
     }
     break;
@@ -691,7 +750,7 @@ static void suboption(void) {
     if (SB_GET() == TELQUAL_SEND) {
       long oospeed, iispeed;
       TerminalSpeeds(&iispeed, &oospeed);
-      netoring.printf("%c%c%c%c%ld,%ld%c%c", IAC, SB, TELOPT_TSPEED, 
+      netoring.xprintf("%c%c%c%c%ld,%ld%c%c", IAC, SB, TELOPT_TSPEED, 
 		      TELQUAL_IS, oospeed, iispeed, IAC, SE);
     }
     break;
@@ -769,7 +828,7 @@ static void suboption(void) {
     if (SB_EOF())
       return;
     if (SB_GET() == TELQUAL_SEND) {
-      const char *dp = env_getvalue("DISPLAY");
+      const char *dp = env_getvalue("DISPLAY", 0);
       if (dp == NULL) {
 	/*
 	 * Something happened, we no longer have a DISPLAY
@@ -778,11 +837,43 @@ static void suboption(void) {
 	send_wont(TELOPT_XDISPLOC, 1);
 	break;
       }
-      netoring.printf("%c%c%c%c%s%c%c", IAC, SB, TELOPT_XDISPLOC,
+      netoring.xprintf("%c%c%c%c%s%c%c", IAC, SB, TELOPT_XDISPLOC,
 		      TELQUAL_IS, dp, IAC, SE);
     }
     break;
     
+#if	defined(AUTHENTICATE)
+  case TELOPT_AUTHENTICATION: {
+    if (!autologin)
+      break;
+    if (SB_EOF())
+      return;
+    switch(SB_GET()) {
+    case TELQUAL_IS:
+      if (my_want_state_is_dont(TELOPT_AUTHENTICATION))
+	return;
+      auth_is(subpointer, SB_LEN());
+      break;
+    case TELQUAL_SEND:
+      if (my_want_state_is_wont(TELOPT_AUTHENTICATION))
+	return;
+      auth_send(subpointer, SB_LEN());
+      break;
+    case TELQUAL_REPLY:
+      if (my_want_state_is_wont(TELOPT_AUTHENTICATION))
+	return;
+      auth_reply(subpointer, SB_LEN());
+      break;
+    case TELQUAL_NAME:
+      if (my_want_state_is_dont(TELOPT_AUTHENTICATION))
+	return;
+      auth_name(subpointer, SB_LEN());
+      break;
+    }
+  }
+  break;
+#endif
+
   default:
     break;
   }
@@ -796,7 +887,7 @@ void lm_will(unsigned char *cmd, int len) {
     return;
   }
   
-  netoring.printf("%c%c%c%c%c%c%c", IAC, SB, TELOPT_LINEMODE, 
+  netoring.xprintf("%c%c%c%c%c%c%c", IAC, SB, TELOPT_LINEMODE, 
 		  DONT, cmd[0], IAC, SE);
 }
 
@@ -813,7 +904,7 @@ void lm_do(unsigned char *cmd, int len) {
     /*@*/	printf("lm_do: no command!!!\n");	/* Should not happen... */
     return;
   }
-  netoring.printf("%c%c%c%c%c%c%c", IAC, SB, TELOPT_LINEMODE, 
+  netoring.xprintf("%c%c%c%c%c%c%c", IAC, SB, TELOPT_LINEMODE, 
 		  WONT, cmd[0], IAC, SE);
 }
 
@@ -836,7 +927,7 @@ void lm_mode(unsigned char *cmd, int len, int init) {
     k |= MODE_ACK;
   }
   
-  netoring.printf("%c%c%c%c%c%c%c", IAC, SB, TELOPT_LINEMODE, LM_MODE,
+  netoring.xprintf("%c%c%c%c%c%c%c", IAC, SB, TELOPT_LINEMODE, LM_MODE,
 		  k, IAC, SE);
   
   setconnmode(0);	/* set changed mode */
@@ -931,11 +1022,11 @@ void slc_mode_import(int def) {
 
 void slc_import(int def) {
   if (def) {
-    netoring.printf("%c%c%c%c%c%c%c%c%c", IAC, SB, TELOPT_LINEMODE,
+    netoring.xprintf("%c%c%c%c%c%c%c%c%c", IAC, SB, TELOPT_LINEMODE,
 		    LM_SLC, 0, SLC_DEFAULT, 0, IAC, SE);
   }
   else {
-    netoring.printf("%c%c%c%c%c%c%c%c%c", IAC, SB, TELOPT_LINEMODE,
+    netoring.xprintf("%c%c%c%c%c%c%c%c%c", IAC, SB, TELOPT_LINEMODE,
 		    LM_SLC, 0, SLC_VARIABLE, 0, IAC, SE);
   }
 }
@@ -1140,6 +1231,7 @@ void env_opt(unsigned char *buf, int len) {
   }
 }
 
+/* OPT_REPLY_SIZE must be a multiple of 2. */
 #define	OPT_REPLY_SIZE	256
 unsigned char *opt_reply;
 unsigned char *opt_replyp;
@@ -1171,6 +1263,7 @@ void env_opt_start_info(void) {
 
 void env_opt_add(const char *ep) {
   const char *vp;
+  const unsigned char *tp;
   unsigned char c;
   
   if (opt_reply == NULL)		/*XXX*/
@@ -1182,12 +1275,13 @@ void env_opt_add(const char *ep) {
     for (ep = env_next(&i,1); ep; ep = env_next(&i,1)) env_opt_add(ep);
     return;
   }
-  vp = env_getvalue(ep);
-  if (opt_replyp + (vp ? strlen(vp) : 0) + strlen(ep) + 6 > opt_replyend)
+  vp = env_getvalue(ep, 1);
+  tp = opt_replyp + (vp ? strlen(vp) * 2 : 0) + strlen(ep) * 2 + 6;
+  if (tp > opt_replyend)
     {
       register int len;
-      opt_replyend += OPT_REPLY_SIZE;
-      len = opt_replyend - opt_reply;
+      len = ((tp - opt_reply) + OPT_REPLY_SIZE - 1) & ~(OPT_REPLY_SIZE - 1);
+      opt_replyend = opt_reply + len;
       opt_reply = (unsigned char *)realloc(opt_reply, len);
       if (opt_reply == NULL) {
 	/*@*/			printf("env_opt_add: realloc() failed!!!\n");
@@ -1722,12 +1816,29 @@ int Scheduler(int block) {
 /*
  * Select from tty and network...
  */
-void telnet(const char * /*user*/) {
+void telnet(const char *user) {
   sys_telnet_init();
   
+#if defined(AUTHENTICATE)
+  {
+    static char local_host[256] = { 0 };
+    int len = sizeof(local_host);
+    
+    if (!local_host[0]) {
+      gethostname(local_host, len);        /* WAS &len!!! */
+      local_host[sizeof(local_host)-1] = 0;
+    }
+    auth_encrypt_init(local_host, hostname, "TELNET", 0);
+    auth_encrypt_user(user);
+  }
+#endif
   
 #if !defined(TN3270)
   if (telnetport) {
+#if	defined(AUTHENTICATE)
+    if (autologin)
+      send_will(TELOPT_AUTHENTICATION, 1);
+#endif
     send_do(TELOPT_SGA, 1);
     send_will(TELOPT_TTYPE, 1);
     send_will(TELOPT_NAWS, 1);
@@ -1736,7 +1847,7 @@ void telnet(const char * /*user*/) {
     send_will(TELOPT_LINEMODE, 1);
     send_will(TELOPT_ENVIRON, 1);
     send_do(TELOPT_STATUS, 1);
-    if (env_getvalue("DISPLAY"))
+    if (env_getvalue("DISPLAY", 0))
       send_will(TELOPT_XDISPLOC, 1);
     if (eight)
       tel_enter_binary(eight);
