@@ -35,7 +35,7 @@
  * From: @(#)sys_term.c	5.16 (Berkeley) 3/22/91
  */
 char st_rcsid[] = 
-  "$Id: sys_term.c,v 1.16 1999/12/12 14:59:45 dholland Exp $";
+  "$Id: sys_term.c,v 1.17 1999/12/17 14:28:47 dholland Exp $";
 
 #include <utmp.h>
 
@@ -58,6 +58,19 @@ int openpty(int *, int *, char *, struct termios *, struct winsize *);
 
 static struct termios termbuf, termbuf2;	/* pty control structure */
 
+#ifdef USE_SSL
+#define DO_SHUTDOWN(x,y) \
+			 if (debug) { \
+			   fprintf(stderr,"Doing SSL_shutdown\n");\
+			   fflush(stderr); \
+			 } \
+			 if (ssl_active_flag) \
+			   SSL_shutdown(ssl_con); \
+			 shutdown((x),(y))
+#else
+#define DO_SHUTDOWN(x,y) shutdown((x),(y))
+#endif
+
 /*static int cleanopen(char *line);*/
 
 /*
@@ -739,6 +752,6 @@ void cleanup(int sig) {
     *p = 'p';
     chmod(line, 0666);
     chown(line, 0, 0);
-    shutdown(net, 2);
-    exit(1);
+    DO_SHUTDOWN(net, 2);
+    exit(0);
 }
